// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//	DOOM graphics stuff for X11, UNIX.
//
//-----------------------------------------------------------------------------

static const char
rcsid[] = "$Id: i_x.c,v 1.6 1997/02/03 22:45:10 b1 Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/time.h>
#include <sys/types.h>

#include <netinet/in.h>
#include <signal.h>


#include <go32.h>
#include <pc.h>
#include <dpmi.h>
#include <dos.h>
#include <sys/nearptr.h>
#include <allegro.h>

#include "doomstat.h"
#include "i_system.h"
#include "multires.h"
#include "m_argv.h"
#include "d_main.h"

#include "doomdef.h"
#include "allegvid.h"

//dosstuff -newly added
int SCREENWIDTH;
int SCREENHEIGHT;
int SCREENPITCH;
int BPP;
int weirdaspect;

char oldkeystate[128];
volatile int mselapsed=0;
int mousepresent;

byte *hicolortable;
short hicolortransmask1,hicolortransmask2;

extern int usejoystick;
extern int usemouse;
int novert=0;

//end of newly added stuff

//
// I_StartFrame
//
void I_StartFrame (void)
  {
  if (usejoystick)
    poll_joystick();
  }

int I_ScanCode2DoomCode (int a)
  {
  switch (a)
    {
    case 0x48: return KEYD_UPARROW;
    case 0x4d: return KEYD_RIGHTARROW;
    case 0x50: return KEYD_DOWNARROW;
    case 0x4b: return KEYD_LEFTARROW;
    case 0x0e: return KEYD_BACKSPACE;
    case 0x2a: return KEYD_RSHIFT;
    case KEY_ALTGR: return KEYD_RALT;
    case KEY_RCONTROL: return KEYD_RCTRL;
    }

  if (key_ascii_table[a]>8)
    return key_ascii_table[a];
  else
    return a+0x80;
  }

int I_DoomCode2ScanCode (int a)
  {
  int i;

  switch (a)
    {
    case KEYD_UPARROW: return 0x48;
    case KEYD_RIGHTARROW: return 0x4d;
    case KEYD_DOWNARROW: return 0x50;
    case KEYD_LEFTARROW: return 0x4b;
    case KEYD_BACKSPACE: return 0x0e;
    }
  if (a>=0x80)
    return (a-0x80);
  else
    {
    for (i=0;i<128;i++)
      if (key_ascii_table[i]==a)
        return i;
    return 0;
    }
  }

void I_GetEvent()
  {
  event_t event;
  int i;

  char keystate[128];
  int xmickeys,ymickeys,buttons;
  static int lastbuttons=0;

  //key presses
  for (i=0;i<128;i++)
    keystate[i]=key[i];
  for (i=0;i<128;i++)
    {
    char normkey,extkey,oldnormkey,oldextkey;

    normkey=keystate[i]&KB_NORMAL; extkey=keystate[i]&KB_EXTENDED;
    oldnormkey=oldkeystate[i]&KB_NORMAL; oldextkey=oldkeystate[i]&KB_EXTENDED;

    if ((normkey!=0)&&(oldnormkey==0))
      {
      event.type=ev_keydown;
      event.data1=I_ScanCode2DoomCode(i);
      D_PostEvent(&event);
      }
    if ((normkey==0)&&(oldnormkey!=0))
      {
      event.type=ev_keyup;
      event.data1=I_ScanCode2DoomCode(i);
      D_PostEvent(&event);
      }
    if ((extkey!=0)&&(oldextkey==0))
      {
//      if ((i==0x48)||(i==0x4d)||(i==0x50)||(i==0x4b)||(i==KEY_ALTGR)||(i==KEY_RCONTROL)||(i=KEY_PRTSCR))
      if ((i==0x48)||(i==0x4d)||(i==0x50)||(i==0x4b)||(i==KEY_ALTGR)||(i==KEY_RCONTROL))
        {
        event.type=ev_keydown;
        event.data1=I_ScanCode2DoomCode(i);
        D_PostEvent(&event);
        }
      else if (i==0x7b)
        {
        event.type=ev_keydown;
        event.data1=KEYD_PAUSE;
        D_PostEvent(&event);
        key[0x7b]=0;break;
        event.type=ev_keyup;
        event.data1=KEYD_PAUSE;
        D_PostEvent(&event);
        }
      }
    if ((extkey==0)&&(oldextkey!=0))
      {
//      if ((i==0x48)||(i==0x4d)||(i==0x50)||(i==0x4b)||(i==KEY_ALTGR)||(i==KEY_RCONTROL)||(i=KEY_PRTSCR))
      if ((i==0x48)||(i==0x4d)||(i==0x50)||(i==0x4b)||(i==KEY_ALTGR)||(i==KEY_RCONTROL))
        {
        event.type=ev_keyup;
        event.data1=I_ScanCode2DoomCode(i);
        D_PostEvent(&event);
        }
      }
    }
  memcpy(oldkeystate,keystate,128);

  //mouse movement
  if ((mousepresent!=-1)&&(usemouse))
    {
    get_mouse_mickeys(&xmickeys,&ymickeys);
    buttons=mouse_b;

    event.type=ev_mouse;
    event.data1=buttons;
    event.data2=xmickeys;
    if (novert==0)
      event.data3=-ymickeys;
    else
      event.data3=0;
    if ((xmickeys!=0)||(ymickeys!=0)||(buttons!=lastbuttons))
      D_PostEvent(&event);
    lastbuttons=buttons;
    }

  //joystick
  if (usejoystick)
    {
    event.type=ev_joystick;
    event.data1=0;
    if (joy_b1) event.data1|=1;
    if (joy_b2) event.data1|=2;
    if (joy_b3) event.data1|=4;
    if (joy_b4) event.data1|=8;
    event.data2=joy_x;
    event.data3=joy_y;
    event.data2=(abs(event.data2)<4)?0:event.data2;
    event.data3=(abs(event.data3)<4)?0:event.data3;
    D_PostEvent(&event);
    }
  }


//
// I_StartTic
//
void I_StartTic()
  {
  I_GetEvent();
  //i dont think i have to do anything else here

  }


//
// I_UpdateNoBlit
//
void I_UpdateNoBlit (void)
{
    // what is this?
}

void I_FinishUpdate(void)
  {
    static int	lasttic;
    int		tics;
    int		i;
    // UNUSED static unsigned char *bigscreen=0;

    // draws little dots on the bottom of the screen
    if (devparm)
    {

	i = I_GetTime();
	tics = i - lasttic;
	lasttic = i;
	if (tics > 20) tics = 20;

	for (i=0 ; i<tics*2 ; i+=2)
       {
	    screens[0][ ((SCREENHEIGHT-1)*SCREENWIDTH + i)*2] = 0xff;
       if (BPP==2)
  	      screens[0][ ((SCREENHEIGHT-1)*SCREENWIDTH + i)*2+1] = 0xff;
       }
	for ( ; i<20*2 ; i+=2)
       {
	    screens[0][ ((SCREENHEIGHT-1)*SCREENWIDTH + i)*2] = 0x0;
       if (BPP==2)
	      screens[0][ ((SCREENHEIGHT-1)*SCREENWIDTH + i)*2+1] = 0x0;
       }
    
    }

  //blast it to the screen
  flipscreens();
  }

//
// I_ReadScreen
//
void I_ReadScreen (byte* scr)
{
    memcpy (scr, screens[0], SCREENWIDTH*SCREENHEIGHT*BPP);
}


void I_SetPalette (byte* palette, int redness)
  {
  int c,i,j;
  PALETTE pal;
  short *tempcolormap;
  byte* tempptr;

  if (BPP==1)
    {
    for (i=0;i<256;i++)
      {
      c=gammatable[usegamma][*palette++];
      pal[i].r=c>>2;
      c=gammatable[usegamma][*palette++];
      pal[i].g=c>>2;
      c=gammatable[usegamma][*palette++];
      pal[i].b=c>>2;
      }
    set_palette(pal);
    }
  else
    {  //replace this with some sorta caching system?
    int tempr[256],tempg[256],tempb[256];

    tempcolormap=(short *)colormaps;
    for (i=0;i<256;i++)
      {
      tempr[i]=gammatable[usegamma][*palette++];
      tempg[i]=gammatable[usegamma][*palette++];
      tempb[i]=gammatable[usegamma][*palette++];
      }
    if ((redness<1)||(redness>8))
      {
      for (i=32;i>0;i--)
        {
        tempptr=hicolortable+256*(i-1);
        for (j=0;j<256;j++)
          {
          *tempcolormap=makecol(tempptr[tempr[j]],tempptr[tempg[j]],tempptr[tempb[j]]);
          tempcolormap++;
          }
        }
      }
    else
      {
      int redconst=redness*31;
      for (i=32;i>0;i--)
        {
        tempptr=hicolortable+256*(i-1)+(256*32*redness);
        for (j=0;j<256;j++)
          {
          *tempcolormap=makecol(redconst+tempptr[tempr[j]],tempptr[tempg[j]],tempptr[tempb[j]]);
          tempcolormap++;
          }
        }
      }
    tempptr=hicolortable+256*32;
    for (i=0;i<256;i++)
      {
      j=tempr[i]*54+tempg[i]*183+tempb[i]*19;  //from the colorspace faq
      *tempcolormap=makecol(255-(j/256),255-(j/256),255-(j/256));  //lookup table?
      tempcolormap[256]=0;
      tempcolormap++;
      pal[i].r=tempr[i]>>2; pal[i].g=tempg[i]>>2; pal[i].b=tempb[i]>>2;
      }
    set_palette(pal);
    }
  }

void I_timer(void)
{
  mselapsed+=5;
  cdcounter+=5;
}
END_OF_FUNCTION(I_timer);

void I_InitGraphics(void)
  {
  static int firsttime=1;
  int i,j;

  if (!firsttime)
    return;
  firsttime=0;

  //init the joystick
  if (usejoystick)
    {
    joy_type = JOY_TYPE_4BUTTON;
    printf("CENTER the joystick and press a key:"); getch(); printf("\n");
    initialise_joystick();
    printf("Push the joystick to the UPPER LEFT corner and press a key:"); getch(); printf("\n");
    calibrate_joystick_tl();
    printf("Push the joystick to the LOWER RIGHT corner and press a key:"); getch(); printf("\n");
    calibrate_joystick_br();
    }

  //enter graphics mode
  enter_graphics_mode();

  //do the hicolorpal table if necessary
  if (BPP==2)
    {
    byte *tempptr,*tempptr2;

    tempptr=hicolortable=(byte *)malloc(256*32*9);
    for (i=0;i<32;i++)
      {
      for (j=0;j<256;j++)
        {
        *tempptr=j*gammatable[3][i*(256/32)]/256;
        tempptr++;
        }
      }
    for (i=1;i<=8;i++)
      {
      tempptr2=hicolortable;
      for (j=0;j<(256*32);j++)
        {
        *tempptr=(byte)(((int)(*tempptr2))*(8-i)/8);
        tempptr++; tempptr2++;
        }
      }
    hicolortransmask1=makecol(127,127,127);
    hicolortransmask2=makecol(63,63,63);
    }

  //init timer
  LOCK_VARIABLE(mselapsed);
  LOCK_FUNCTION(I_timer);
  install_timer();
  install_int(I_timer,5);

  //init the mouse
  if (usemouse)
    {
    mousepresent=install_mouse();
    if (mousepresent!=-1)
      show_mouse(NULL);
    if (M_CheckParm("-novert"))
      novert=1;
    }

  //init keyboard
  memset(oldkeystate,0,128);
  install_keyboard();
  }

void I_ShutdownGraphics(void)
{
//  if (KB_CAPSLOCK_FLAG)
//    set_leds(KB_CAPSLOCK_FLAG);
  remove_keyboard();
  if (mousepresent!=-1)
    remove_mouse();
  remove_timer();
  set_gfx_mode(GFX_TEXT,80,25,0,0);
}


